module NumRu
  module DCLExt
    module Anim

      class Stream

        include Anim

        DEPTH = 24
        BPP = 24

        DEFAULT_FRAMERATE = 4
        DEFAULT_FILENAME = "dcl.mp4".freeze
        DEFAULT_X264 = {
          "pass" => "qual",
          "quantizer" => 18,
          "qp-min" => 10,
          "qp-max" => 25,
          "me" => "umh",
          "subme" => 6,
          "ref" => 3,
          "threads" => 0,
          "cabac" => "true",
          "dct8x8" => "false"
        }.freeze



        attr_accessor :filename, :framerate, :x264
        attr_reader :pipe
        attr_reader :buffersize

        def initialize(pipe, opts={})
          @pipe = pipe
          @buffersize = width * height * DEPTH
          @num_frames = 0
          @source_id = nil
          @framerate = opts["framerate"] || DEFAULT_FRAMERATE
          @filename = opts["filename"] || DEFAULT_FILENAME
          @x264 = DEFAULT_X264.dup
          if x264 = opts["x264"]
            case x264
            when Hash
              # nothing to do
            when String # each option must be separated by space e.g. "key1=val key2=val key3=val"
              x264 = x264.split(" ").map{|c| c.split("=")}.inject({}){|h,a| h[a[0]]=a[1];h}
            else
              raise "x264 must be Hash or String"
            end
            @x264.update(x264)
          end
        end

        def start
          require "gst"

          Gst.init

          loop = GLib::MainLoop.new(nil, false)

          x264opts = @x264.map{|k,v| "#{k}=#{v}"}.join(" ")
          cmd = "appsrc min-latency=100000000 block=true name=dclsource caps=\"#{caps_str}\" ! queue ! videoconvert ! video/x-raw,format=(string)I420,width=#{width},height=#{height} ! queue ! x264enc #{x264opts} ! queue ! mp4mux ! filesink location=#{@filename}"

          bin, error = Gst.parse_launch(cmd)
          if bin.nil?
            raise error.message
          end

          appsrc = bin.get_child("dclsource")
          appsrc.format = Gst::Format::TIME
          appsrc.max_bytes = height*width*DEPTH*5

          bus = bin.bus
          bus.add_watch do |bus, message|
            case message.type
            when Gst::Message::EOS
              loop.quit
            when Gst::Message::ERROR
              p message.parse
              loop.quit
            end
            true
          end

          appsrc.signal_connect('need-data') do |elm, length|
            start_feed(elm)
          end
          appsrc.signal_connect('enough-data') do |elm|
            stop_feed
          end

          bin.play
          begin
            loop.run
          rescue Interrupt
          ensure
            @pipe.close
            bin.stop
          end
        end

        def caps_str
          "video/x-raw,format=(string)RGB,width=(int)#{width},height=(int)#{height},depth=(int)#{DEPTH},bpp=(int)#{BPP},framerate=#{@framerate}/1,endianness=4321,red_mask=16711680,green_mask=65280,blue_mask=255"
        end

        def get_data
          sleep(0.1)
          @pipe.read(@buffersize/8)
        end

        def start_feed(appsrc)
          unless @source_id
            @source_id = GLib::Idle.add do
              if data = get_data
                buffer = Gst::Buffer.new(data.size*8)
                buffer.caps = Gst::Caps.parse(caps_str)
                buffer.duration = 10**9/@framerate
                buffer.timestamp = buffer.duration * @num_frames
                buffer.set_data(data)
#                ret = appsrc.signal_emit("push-buffer", buffer)
                ret = appsrc.push_buffer(buffer)
                if ret.name == "GST_FLOW_OK"
                  @num_frames += 1
                  true
                else
                  p ret.name
                  appsrc.end_of_stream
                  false
                end
              else
                appsrc.end_of_stream
                false
              end
            end
          end
        end
        def stop_feed
          if @source_id
            GLib::Source.remove(@source_id)
            @source_id = nil
          end
        end

      end # class Stream

    end # module Anim
  end # module DCLExt

end # module NumRu
